package com.persagy.proxy.adm.service.impl;

import java.time.format.DateTimeFormatter;
import java.util.*;

import cn.hutool.core.collection.CollUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.persagy.dmp.common.context.AppContext;
import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.metadata.OrderItem;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.google.common.collect.Lists;
import com.persagy.dmp.basic.dto.RequestData;
import com.persagy.dmp.basic.model.QueryCriteria;
import com.persagy.dmp.common.helper.SpringHelper;
import com.persagy.dmp.common.model.response.CommonResult;
import com.persagy.dmp.define.client.DigitalGraphDefineFacade;
import com.persagy.dmp.define.client.DigitalRelationDefineFacade;
import com.persagy.dmp.define.entity.GraphDefine;
import com.persagy.dmp.define.entity.RelationDefine;
import com.persagy.dmp.digital.client.DigitalObjectFacade;
import com.persagy.dmp.digital.client.DigitalRelationFacade;
import com.persagy.dmp.digital.client.DigitalRelationProjectCalFacade;
import com.persagy.dmp.digital.entity.ObjectRelation;
import com.persagy.dmp.digital.entity.ObjectRelationProjectCal;
import com.persagy.dmp.simple.client.DigitalObjectSimpleFacade;
import com.persagy.proxy.adm.constant.AdmCommonConstant;
import com.persagy.proxy.adm.constant.AdmRelationTypeEnum;
import com.persagy.proxy.adm.constant.ObjTypeMapping;
import com.persagy.proxy.adm.service.IRelationReportService;
import com.persagy.proxy.adm.strategy.relationdata.RelationObjectStrategy;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateTime;
import cn.hutool.core.util.IdUtil;
import cn.hutool.core.util.StrUtil;

/**
 * 报表/模板下载
 * 
 * @version 1.0.0
 * @company persagy 
 * @author zhangqiankun
 * @date 2021年8月31日 下午4:43:15
 */
@Service
public class RelationReportService implements IRelationReportService {

	@Value("${middleware.group.code}")
	private String defaultCode;

	@Override
	public List<Map<String, Object>> queryObjectCountByObjInfos(String groupCode, Set<String> classCodes, Set<String> objTypes, List<String> groups) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		RequestData requestData = new RequestData();
		requestData.setClassCodes(classCodes);
		requestData.setObjTypes(objTypes);
		requestData.setGroups(groups);
		List<Map<String,Object>> list = DigitalObjectSimpleFacade.queryObjectCountByObjInfos(groupCode, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, requestData);
		return CollectionUtil.isEmpty(list) ? Lists.newArrayList() : list;
	}
	
	@Override
	public List<GraphDefine> getGraphDefine(String groupCode, String projectId, String graphCode) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		if (StrUtil.isNotBlank(graphCode)) {
			criteria.put("code", graphCode);
		}
		queryCriteria.setCriteria(criteria);
		List<GraphDefine> graphDefines = DigitalGraphDefineFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return graphDefines;
	}
	
	@Override
	public List<RelationDefine> getRelationDefine(String groupCode, String projectId, String graphCode, String relCode) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		if (StrUtil.isNotBlank(graphCode)) {
			criteria.put("graphCode", graphCode);
		}
		if (StrUtil.isNotBlank(relCode)) {
			criteria.put("code", relCode);
		}
		queryCriteria.setCriteria(criteria);
		List<RelationDefine> relationDefines = DigitalRelationDefineFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return relationDefines;
	}
	
	@Override
	public ObjectNode getRelationProjectCal(String groupCode, String projectId, String relType) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		AdmRelationTypeEnum relationTypeEnum = AdmCommonConstant.RELATION_TYPE_MAP.get(relType);
		if (relationTypeEnum == null) {
			return null;
		}
		
		ObjectNode objectNode = this.getRelationProjectCal(groupCode, projectId, relationTypeEnum.getGraphCode(), relationTypeEnum.getRelCode(), null);
		if (objectNode != null) {
			String zoneType = objectNode.get("zoneType") == null ? null : objectNode.get("zoneType").asText();
			long count = this.countRelationObject(groupCode, projectId, relationTypeEnum.getGraphCode(), relationTypeEnum.getRelCode(), zoneType);
			objectNode.put("count", count);
			objectNode.put("relationType", relType);
			
			// 获取此图类型和边类型对应的数据记录
			List<RelationDefine> relationDefines = this.getRelationDefine(groupCode, projectId, relationTypeEnum.getGraphCode(), relationTypeEnum.getRelCode());
			if (CollectionUtil.isNotEmpty(relationDefines)) {
				RelationDefine relationDefine = relationDefines.get(0);
				objectNode.put("conneObject", relationDefine.getTargetObjs());
				objectNode.put("relationTypeCode", relationDefine.getCode());
				objectNode.put("relationTypeName", relationDefine.getName());
				objectNode.put("automatic", (null == relationDefine.getAutomaticFlag() || relationDefine.getAutomaticFlag()) ? false : true);
			}
			
			String prefix = ObjTypeMapping.getRelCodePrefix(relationTypeEnum.getRelCode());
			String suffix = ObjTypeMapping.getRelCodeSuffix(relationTypeEnum.getRelCode());
			objectNode.put("mainObject", ObjTypeMapping.PREFIX_NAME.get(prefix));
			objectNode.put("fromObject", ObjTypeMapping.PREFIX_NAME.get(suffix));
			objectNode.putPOJO("statistics", new JSONObject());
		}
		
		return objectNode;
	}
	
	@Override
	public ObjectNode getRelationProjectCal(String groupCode, String projectId, String graphCode, String relCode, String zoneType) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put("graphCode", graphCode);
		criteria.put("relCode", relCode);
		if (StrUtil.isNotBlank(zoneType)) {
			criteria.put("zoneType", zoneType);
		}
		queryCriteria.setCriteria(criteria);
		List<ObjectNode> objectNodes = DigitalRelationProjectCalFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		if (CollectionUtil.isEmpty(objectNodes)) {
			return null;
		}
		return objectNodes.get(0);
	}
	
	@Override
	public long countRelationObject(String groupCode, String projectId, String graphCode, String relCode, String zoneType) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put("graphCode", graphCode);
		criteria.put("relCode", relCode);
		if (StrUtil.isNotBlank(zoneType)) {
			criteria.put("relValue", zoneType);
		}
		queryCriteria.setCriteria(criteria);
		queryCriteria.setOnlyCount(true);
		CommonResult<List<ObjectRelation>> result = DigitalRelationFacade.queryPrototype(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return result == null ? 0 : (result.getCount() == null ? 0 : result.getCount());
	}
	
	@Override
	public long countObject(String groupCode, String projectId, String classCode, String objType) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put("classCode", classCode);
		if (StrUtil.isNotBlank(objType)) {
			criteria.put("objType", objType);
		}
		queryCriteria.setCriteria(criteria);
		queryCriteria.setOnlyCount(true);
		CommonResult<List<ObjectNode>> result = DigitalObjectFacade.queryPrototype(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return result == null ? 0 : (result.getCount() == null ? 0 : result.getCount());
	}
	
	@Override
	public ObjectNode getObjectNode(String groupCode, String projectId, String id) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put("id", id);
		queryCriteria.setCriteria(criteria);
		List<ObjectNode> objectNodes = DigitalObjectFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return CollectionUtil.isEmpty(objectNodes) ? objectMapper.createObjectNode() : objectNodes.get(0); 
	}
	
	@Override
	public List<ObjectNode> findObjectNodes(String groupCode, String projectId, String objType, String code, String value) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		if (StrUtil.isBlank(code)) {
			return null;
		}
		String key = RelationObjectStrategy.OBJNAME_2_TYPE.get(code);
		if (StrUtil.isBlank(key)) {
			return null;
		}
		
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put(key, value);
		criteria.put("objType", objType);
		queryCriteria.setCriteria(criteria);
		List<ObjectNode> objectNodes = DigitalObjectFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return objectNodes;
	}
	
	@Override
	public ObjectNode getObjectByCodeAndId(String groupCode, String projectId, String graphCode, String relCode, String objFrom, String objTo) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		// 获取出关联对象
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put("graphCode", graphCode);
		criteria.put("relCode", relCode);
		criteria.put("objFrom", objFrom);
		criteria.put("objTo", objTo);
		queryCriteria.setCriteria(criteria);
		List<ObjectRelation> objectRelations = DigitalRelationFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		if (CollectionUtil.isEmpty(objectRelations)) {
			return objectMapper.createObjectNode();
		}
		
		// 获取关联对象的本地名称
		ObjectRelation objectRelation = objectRelations.get(0);
		String relationId = StrUtil.isBlank(objFrom) ? objectRelation.getObjFrom() : objectRelation.getObjTo();
		return this.getObjectNode(groupCode, projectId, relationId); 
	}
	
	@Override
	public List<ObjectRelation> findObjectRelationList(String groupCode, String projectId, String graphCode, String relCode) {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		
		// 获取出关联对象
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		
		QueryCriteria queryCriteria = new QueryCriteria();
		ObjectNode criteria = objectMapper.createObjectNode();
		criteria.put("graphCode", graphCode);
		criteria.put("relCode", relCode);
		queryCriteria.setCriteria(criteria);
		List<ObjectRelation> objectRelations = DigitalRelationFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		return CollectionUtil.isEmpty(objectRelations) ? Lists.newArrayList() : objectRelations;
	}
	/**
	 * 初始化项目关系计算数据
	 *
	 * @param groupCode: 集团编码
	 * @param projectId: 项目id
	 * @return
	 */
	@Override
	public boolean createRelationProjectCal(String groupCode, String projectId) throws JsonProcessingException {
		groupCode = StrUtil.isBlank(groupCode) ? defaultCode : groupCode;
		// 1.验证是否已经添加过
		QueryCriteria queryCriteria = new QueryCriteria();
		queryCriteria.setOnlyCount(true);
		CommonResult<List<ObjectNode>> prototype = DigitalRelationProjectCalFacade.queryPrototype(groupCode, projectId,
				AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, queryCriteria);
		if (prototype != null && prototype.getCount() != null && prototype.getCount() > 0) {
			return true;
		}
		
		// 2.获取所有的边类型
		QueryCriteria criteria = new QueryCriteria();
		List<RelationDefine> list = DigitalRelationDefineFacade.query(groupCode, projectId,
				AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, criteria);
		if (CollectionUtil.isEmpty(list)) {
			return true;
		}
		
		// 3.组装关系计算数据
		List<ObjectNode> nodes = new ArrayList<ObjectNode>();
		ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
		for (RelationDefine relationDefine : list) {
			// 3.1 判断是否需要加入项目
			if (null==relationDefine.getCalFlag()
					|| !relationDefine.getCalFlag()){
				continue;
			}
			// 3.2 判断zoneTypes是否为多个,如果多个则根据zoneType创建
			if (CollUtil.isNotEmpty(relationDefine.getZoneTypes())){
				for (String zoneType : relationDefine.getZoneTypes()) {
					nodes.add(createRelationProjectCalInternal(objectMapper,groupCode,projectId,relationDefine,zoneType));
				}
				continue;
			}
			// 3.3 如果没有分区则直接创建一个
			nodes.add(createRelationProjectCalInternal(objectMapper,groupCode,projectId,relationDefine,null));
		}
		List<ObjectRelationProjectCal> result = DigitalRelationProjectCalFacade.create(groupCode, projectId,
				AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, nodes);
		return CollectionUtil.isNotEmpty(result);
	}
	/***
	 * Description: 创建存储的对象
	 * @param groupCode : 集团编码
	 * @param projectId : 项目id
	 * @param relationDefine : 关系定义
	 * @param zoneType : 空间分区类型
	 * @return : void
	 * @author : lijie
	 * @date :2021/10/14 16:14
	 * Update By lijie 2021/10/14 16:14
	 */
	private ObjectNode createRelationProjectCalInternal(ObjectMapper objectMapper, String groupCode,
												  		String projectId, RelationDefine relationDefine,
														String zoneType) throws JsonProcessingException {
		ObjectNode objectNode = objectMapper.createObjectNode();
		// 缺少业务分区类型
		objectNode.put("groupCode", groupCode);
		objectNode.put("projectId", projectId);
		objectNode.put("id", IdWorker.getIdStr());
		objectNode.put("graphCode", relationDefine.getGraphCode());
		objectNode.put("relCodeName", relationDefine.getName());
		objectNode.put("relCode", relationDefine.getCode());
		objectNode.put("sourceFlag", relationDefine.getSourceFlag());	// 是否需要配置源末端
		objectNode.put("automaticFlag", relationDefine.getAutomaticFlag());	// 是否自动计算类型
		objectNode.put("manual", relationDefine.getManual());	// 手动计算类型 1 禁用  2提示引导 3手动维护
		objectNode.put("computationalState", 1);	// 计算状态 1初始状态 2需要计算 3计算中 4等待中 5计算失败（如果处于计算中状态，则不再次调用）
		objectNode.put("zoneType", zoneType);	// 业务分区类型
		objectNode.put("systemType", relationDefine.getSystemType());	// 风系统1,水系统0
		objectNode.put("computeVersion", 0);	// 计算版本
		objectNode.put("mepSystemType", CollUtil.isNotEmpty(relationDefine.getMepSystemType())
						?objectMapper.writeValueAsString(relationDefine.getMepSystemType()):null);	// 机电系统类型
		objectNode.put("creator", AdmCommonConstant.USER_ID);	// 创建人
		objectNode.put("modifier", AdmCommonConstant.USER_ID);	// 最后修改人
		return objectNode;
	}

	/**
	 * 查询关系总览（空表，先插入）
	 * @param content: 模糊查询卡片名称关键字
	 */
	@Override
	public JSONObject overview(String content) throws JsonProcessingException {
		JSONObject result = new JSONObject();
		if (StrUtil.isBlank(content)) {
			this.createRelationProjectCal(AppContext.getContext().getGroupCode(), AppContext.getContext().getProjectId());
		}
		List<JSONObject> parentList = new ArrayList<JSONObject>();
		// 获取树结构
		this.getGraphRelationProjectTree(parentList, AppContext.getContext().getGroupCode(),
				AppContext.getContext().getProjectId(), content);
		result.put("content", parentList);
		result.put("message", "");
		result.put("pageNumber", 1);
		result.put("pageSize", 1000);
		result.put("result", "success");
		result.put("total", parentList.size());
		return result;
	}
	
	/**
	 * 获取图类型下的边类型的项目计算关系数据，并形成三级树结构
	 * 
	 * @param parentList
	 * @param groupCode
	 * @param projectId
	 * @param content
	 */
	private void getGraphRelationProjectTree(List<JSONObject> parentList, String groupCode, String projectId, String content) {
		boolean isQuery = StringUtils.isNotBlank(content);
		// 获取所有的一级图类型定义
		QueryCriteria parentCriteria = new QueryCriteria();
		List<GraphDefine> topList = DigitalGraphDefineFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, parentCriteria);
		for (int i = topList.size() - 1; i >= 0; i--) {
			// 获取所有的二级图类型定义
			QueryCriteria childCriteria = new QueryCriteria();
			List<GraphDefine> childList = DigitalGraphDefineFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, childCriteria);
			if (CollectionUtil.isEmpty(childList)) {
				if (isQuery) {
					topList.remove(i);		//移除
				}
				continue;
			}
			
			GraphDefine topGraph = topList.get(i);
			JSONObject parent = this.transfer(topGraph);
			// 二级数据封装
			for (int j = childList.size() - 1; j >= 0; j--) {
				GraphDefine childGraph = topList.get(i);
				JSONObject child = this.transfer(childGraph);
				// 查询出对应项目关系计算数据
				QueryCriteria projectCriteria = new QueryCriteria();
				ObjectMapper objectMapper = SpringHelper.getBean(ObjectMapper.class);
				ObjectNode criteria = objectMapper.createObjectNode();
				criteria.put("projectId", projectId);
				if (isQuery) {
					ObjectNode condition = objectMapper.createObjectNode();
					condition.put("$like", content);
					criteria.putPOJO("name", condition); // 模糊查询
				}
				projectCriteria.setCriteria(criteria);
				projectCriteria.setOrders(Collections.singletonList(OrderItem.asc("relCode")));
				List<ObjectNode> projectList = DigitalRelationProjectCalFacade.query(groupCode, projectId, AdmCommonConstant.APP_ID, AdmCommonConstant.USER_ID, projectCriteria);
				if (CollectionUtil.isEmpty(projectList)) {
					if (isQuery) {
						childList.remove(i);		//移除
					}
					continue;
				}
				// 三级数据封装
				for (ObjectNode objectNode : projectList) {
					this.transfer(objectNode, childGraph.getId());
				}
				child.put("relationTypeProjectList", projectList);
			}
			// 再判断一次，防止当二级有数据但被清除了的情况下，一级还会存在
			if (isQuery && CollectionUtil.isEmpty(childList)) {
				childList.remove(i);		//移除
				continue;
			}
			parent.put("childGraphicTypeList", childList);
			parentList.add(parent);
		}
	}
	
	/**
	 * 获取转换后的数据，图类型数据转换
	 * 
	 * @param graphDefine
	 * @return
	 */
	private JSONObject transfer(GraphDefine graphDefine) {
		JSONObject top = new JSONObject();
		top.put("sort", 1);
		top.put("note", "");
		top.put("statistics", Lists.newArrayList());
		top.put("graphTypeName", graphDefine.getName());
		top.put("graphTypeId", graphDefine.getId());
		top.put("graphTypeCode", graphDefine.getCode());
		top.put("lastUpdate", graphDefine.getModifiedTime() == null ? null : graphDefine.getModifiedTime().toString(DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")));
		top.put("createTime", graphDefine.getCreationTime() == null ? null : graphDefine.getCreationTime().toString(DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")));
		return top;
	}
	
	/**
	 * 获取转换后的数据，项目计算数据转换
	 * 
	 * @param relationProject
	 * @return
	 */
	private void transfer(ObjectNode relationProject, String graphicId) {
		String relCode = relationProject.get("relCode") == null ? null : relationProject.get("relCode").asText();
		
		relationProject.put("graphicId", graphicId);
		relationProject.putPOJO("statistics", Lists.newArrayList());
		relationProject.put("automatic", "1".equals(relationProject.get("automaticFlag") == null ? null : relationProject.get("automaticFlag").asText()));
		relationProject.put("source", "1".equals(relationProject.get("sourceFlag") == null ? null : relationProject.get("sourceFlag").asText()));
		relationProject.put("conneObject", ObjTypeMapping.getConneObject(relCode));
		
		Long modifiedTime = relationProject.get("modifiedTime") == null ? null : relationProject.get("modifiedTime").asLong();
		Long creationTime = relationProject.get("creationTime") == null ? null : relationProject.get("creationTime").asLong();
		
		relationProject.put("lastUpdate", modifiedTime == null ? null : new DateTime(modifiedTime).toString());
		relationProject.put("createTime", creationTime == null ? null : new DateTime(creationTime).toString());
	}
	
}
